/*
 * Copyright (c) 2007-2009-2010 the original author or authors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.as3commons.reflect {

	import flash.system.ApplicationDomain;
	import flash.utils.Dictionary;
	
	import org.as3commons.lang.ClassNotFoundError;
	import org.as3commons.lang.ClassUtils;
	import org.as3commons.lang.HashArray;
	import org.as3commons.logging.ILogger;
	import org.as3commons.logging.LoggerFactory;
	
	/**
	 * Provides information about the characteristics of a class or an interface.
	 * These are the methods, accessors (getter/setter), variables and constants,
	 * but also if the class is <code>dynamic</code> and <code>final</code>.
	 *
	 * <p>Note that information about an object cannot be retrieved by calling the
	 * <code>Type</code> constructor. Instead use one of the following static
	 * methods:</p>
	 *
	 * <p>In case of an instance:
	 * <code>var type:Type = Type.forInstance(myInstance);</code>
	 * </p>
	 *
	 * <p>In case of a <code>Class</code> variable:
	 * <code>var type:Type = Type.forClass(MyClass);</code>
	 * </p>
	 *
	 * <p>In case of a classname:
	 * <code>var type:Type = Type.forName("MyClass");</code>
	 * </p>
	 *
	 * @author Christophe Herreman
	 * @author Martino Piccinato
	 * @author Andrew Lewisohn
	 */
	public class Type extends MetaDataContainer {

		public static const UNTYPED:Type = new Type(ApplicationDomain.currentDomain);
		{
			UNTYPED.fullName = '*';
			UNTYPED.name = '*';
		}

		public static const VOID:Type = new Type(ApplicationDomain.currentDomain);
		{
			VOID.fullName = 'void';
			VOID.name = 'void';
		}

		public static const PRIVATE:Type = new Type(ApplicationDomain.currentDomain);
		{
			PRIVATE.fullName = 'private';
			PRIVATE.name = 'private';
		}

		private static const MEMBER_PROPERTY_NAME:String = 'name';

		private static var logger:ILogger = LoggerFactory.getClassLogger(Type);

		private static var typeProvider:ITypeProvider;
		
		/**
		 * type need cache
		 */ 
		public static var classTypeCache_enable:Boolean = false
			
		/**
		 * filter package in type cache
		 */ 
		public static var filter_packages:Array = [];
			

		// --------------------------------------------------------------------
		//
		// Static Methods
		//
		// --------------------------------------------------------------------

		/**
		 * Returns a <code>Type</code> object that describes the given instance.
		 *
		 * @param instance the instance from which to get a type description
		 */
		public static function forInstance(instance:*, applicationDomain:ApplicationDomain = null, fullyQualifiedClassName:String = null):Type {
			applicationDomain = (applicationDomain == null) ? ApplicationDomain.currentDomain : applicationDomain;
			var result:Type;
			var clazz:Class = org.as3commons.lang.ClassUtils.forInstance(instance, applicationDomain);

			if (clazz != null) 
			{
				var fullyQualifiedClassName:String;
				
				if(fullyQualifiedClassName == null ){
					fullyQualifiedClassName = org.as3commons.lang.ClassUtils.getFullyQualifiedName(clazz);
				}
				
				result = Type.forClass(clazz, applicationDomain,fullyQualifiedClassName);
			}
			return result;
		}

		/**
		 * Returns a <code>Type</code> object that describes the given classname.
		 *
		 * @param name the classname from which to get a type description
		 */
		public static function forName(name:String, applicationDomain:ApplicationDomain = null):Type {
			applicationDomain = (applicationDomain == null) ? ApplicationDomain.currentDomain : applicationDomain;
			var result:Type;

			/*if(name.indexOf("$")!=-1){
			   return Type.PRIVATE;
			 }*/
			switch (name) {
				case "void":
					result = Type.VOID;
					break;
				case "*":
					result = Type.UNTYPED;
					break;
				default:
					try {
						if (getTypeProvider().getTypeCache().contains(name)) {
							result = getTypeProvider().getTypeCache().get(name);
						} else {
							result = Type.forClass(org.as3commons.lang.ClassUtils.forName(name, applicationDomain), applicationDomain);
						}
					} catch (e:ReferenceError) {
						logger.warn("Type.forName error: " + e.message + " The class '" + name + "' is probably an internal class or it may not have been compiled.");
					} catch (e:ClassNotFoundError) {
						logger.warn("The class with the name '{0}' could not be found in the application domain '{1}'", name, applicationDomain);
					}
			}
			return result;
		}

		/**
		 * Returns a <code>Type</code> object that describes the given class.
		 *
		 * @param clazz the class from which to get a type description
		 */
		public static function forClass(clazz:Class, applicationDomain:ApplicationDomain = null , fullyQualifiedClassName:String = null):Type 
		{
			
			applicationDomain = (applicationDomain == null) ? ApplicationDomain.currentDomain : applicationDomain;
			
			var result:Type;
			
			if(fullyQualifiedClassName==null){
				
				fullyQualifiedClassName = org.as3commons.lang.ClassUtils.getFullyQualifiedName(clazz);
			}
						
			if(getTypeProvider().getTypeCache().contains(fullyQualifiedClassName))
			{
				result = getTypeProvider().getTypeCache().get(fullyQualifiedClassName);
			}
			else
			{
				result = getTypeProvider().getType(clazz, applicationDomain,fullyQualifiedClassName);
			}
						
			return result;
		}
		
		private static function checkNeedCache(fullyQualifiedClassName:String):Boolean
		{
			//filter_package package need cache
			if( filter_packages.length >0 )
			{
				for each(var pg:String in filter_packages){
					if(pg!=null){
						if(fullyQualifiedClassName.indexOf(pg) != -1){
							return true;
						}
					}
				}
			} 
			
			//all interface need cache
			if (getTypeProvider().getTypeCache().contains(fullyQualifiedClassName)) 
			{
				if(Type( getTypeProvider().getTypeCache().get(fullyQualifiedClassName) ).isInterface){
					return true;
				}
				
			}
			
			return classTypeCache_enable;
		}

		/**
		 *
		 */
		public static function getTypeProvider():ITypeProvider {
			if (typeProvider == null) {
				typeProvider = new XmlTypeProvider();
			}
			return typeProvider;
		}

		// --------------------------------------------------------------------
		//
		// Constructor
		//
		// --------------------------------------------------------------------

		/**
		 * Creates a new <code>Type</code> instance.
		 */
		public function Type(applicationDomain:ApplicationDomain) {
			super();
			initType(applicationDomain);
		}

		/**
		 * Initializes the current <code>Type</code> instance.
		 */
		protected function initType(applicationDomain:ApplicationDomain):void {
			_accessors = [];
			_staticConstants = [];
			_constants = [];
			_staticVariables = [];
			_variables = [];
			_extendsClasses = [];
			_interfaces = [];
			_applicationDomain = applicationDomain;
		}
		
		/**
		 * dispose
		 */ 
		public function dispose():void
		{
			if(!checkNeedCache(fullName))
			{
				clazz = null;
				_accessors = null
				_staticConstants = null
				_constants = null
				_staticVariables =null
				_variables = null
				_extendsClasses = null
				_interfaces = null
				_applicationDomain = null
				if(_fields!=null) _fields.dispose();
				_fields = null ;
				if(_methods!=null) _methods.dispose();
				_methods = null;
				
				
				getTypeProvider().clearByKey(fullName);
			}
		}
		
		// --------------------------------------------------------------------
		//
		// Properties
		//
		// --------------------------------------------------------------------

		// ----------------------------
		// applicationDomain
		// ----------------------------

		private var _applicationDomain:ApplicationDomain;

		/**
		 * The ApplicationDomain which is able to retrieve the object definition for this type. The definition does not
		 * necessarily have to be part of this <code>ApplicationDomain</code>, it could possibly be present in the parent domain as well.
		 */
		public function get applicationDomain():ApplicationDomain {
			return _applicationDomain;
		}

		/**
		 * @private
		 */
		public function set applicationDomain(value:ApplicationDomain):void {
			_applicationDomain = value;
		}

		// ----------------------------
		// name
		// ----------------------------

		private var _alias:String;

		public function get alias():String {
			return _alias;
		}

		public function set alias(value:String):void {
			_alias = value;
		}

		// ----------------------------
		// name
		// ----------------------------

		private var _name:String;

		/**
		 * The name of the type
		 */
		public function get name():String {
			return _name;
		}

		/**
		 * @private
		 */
		public function set name(value:String):void {
			_name = value;
		}

		// ----------------------------
		// fullName
		// ----------------------------

		private var _fullName:String;

		/**
		 * The fully qualified name of the type, this includes the namespace
		 */
		public function get fullName():String {
			return _fullName;
		}

		/**
		 * @private
		 */
		public function set fullName(value:String):void {
			_fullName = value;
		}

		// ----------------------------
		// clazz
		// ----------------------------

		private var _class:Class;

		/**
		 * The Class of the <code>Type</code>
		 */
		public function get clazz():Class {
			return _class;
		}

		/**
		 * @private
		 */
		public function set clazz(value:Class):void {
			_class = value;
		}

		// ----------------------------
		// isDynamic
		// ----------------------------

		private var _isDynamic:Boolean;

		/**
		 * True if the <code>Type</code> is dynamic
		 */
		public function get isDynamic():Boolean {
			return _isDynamic;
		}

		/**
		 * @private
		 */
		public function set isDynamic(value:Boolean):void {
			_isDynamic = value;
		}

		// ----------------------------
		// isFinal
		// ----------------------------

		private var _isFinal:Boolean;

		/**
		 * True if the <code>Type</code> is final
		 */
		public function get isFinal():Boolean {
			return _isFinal;
		}

		/**
		 * @private
		 */
		public function set isFinal(value:Boolean):void {
			_isFinal = value;
		}

		// ----------------------------
		// isStatic
		// ----------------------------

		private var _isStatic:Boolean;

		/**
		 * True if the <code>Type</code> is static
		 */
		public function get isStatic():Boolean {
			return _isStatic;
		}

		/**
		 * @private
		 */
		public function set isStatic(value:Boolean):void {
			_isStatic = value;
		}

		// ----------------------------
		// isInterface
		// ----------------------------

		private var _isInterface:Boolean;

		/**
		 * True if the <code>Type</code> is an interface
		 */
		public function get isInterface():Boolean {
			return _isInterface;
		}

		/**
		 * @private
		 */
		public function set isInterface(value:Boolean):void {
			_isInterface = value;
		}

		// ----------------------------
		// interfaces
		// ----------------------------

		private var _interfaces:Array;

		public function get interfaces():Array {
			return _interfaces;
		}

		/**
		 * @private
		 */
		public function set interfaces(value:Array):void {
			_interfaces = value;
		}

		// ----------------------------
		// constructor
		// ----------------------------

		private var _constructor:Constructor;

		/**
		 * A reference to a <code>Constructor</code> instance that describes the constructor of the <code>Type</code>
		 * @see org.as3commons.reflect.Constructor Constructor
		 */
		public function get constructor():Constructor {
			return _constructor;
		}

		/**
		 * @private
		 */
		public function set constructor(constructor:Constructor):void {
			_constructor = constructor;
		}

		// ----------------------------
		// accessors
		// ----------------------------

		private var _accessors:Array;

		/**
		 * An array of <code>Accessor</code> instances
		 * @see org.as3commons.reflect.Accessor Accessor
		 */
		public function get accessors():Array {
			return _accessors;
		}

		/**
		 * @private
		 */
		public function set accessors(value:Array):void {
			_accessors = value;
		}

		// ----------------------------
		// methods
		// ----------------------------

		private var _methods:HashArray;

		/**
		 * An array of <code>Method</code> instances
		 * @see org.as3commons.reflect.Method Method
		 */
		public function get methods():Array {
			return (_methods != null) ? _methods.getArray() : [];
		}

		/**
		 * @private
		 */
		public function set methods(value:Array):void {
			_methods = new HashArray(MEMBER_PROPERTY_NAME, false, value);
		}

		// ----------------------------
		// staticConstants
		// ----------------------------

		private var _staticConstants:Array;

		/**
		 * An array of <code>IMember</code> instances that describe the static constants of the <code>Type</code>
		 * @see org.as3commons.reflect.IMember IMember
		 */
		public function get staticConstants():Array {
			return _staticConstants;
		}

		/**
		 * @private
		 */
		public function set staticConstants(value:Array):void {
			_staticConstants = value;
		}

		// ----------------------------
		// constants
		// ----------------------------

		private var _constants:Array;

		/**
		 * An array of <code>IMember</code> instances that describe the constants of the <code>Type</code>
		 * @see org.as3commons.reflect.IMember IMember
		 */
		public function get constants():Array {
			return _constants;
		}

		/**
		 * @private
		 */
		public function set constants(value:Array):void {
			_constants = value;
		}

		// ----------------------------
		// staticVariables
		// ----------------------------

		private var _staticVariables:Array;

		/**
		 * An array of <code>IMember</code> instances that describe the static variables of the <code>Type</code>
		 * @see org.as3commons.reflect.IMember IMember
		 */
		public function get staticVariables():Array {
			return _staticVariables;
		}

		/**
		 * @private
		 */
		public function set staticVariables(value:Array):void {
			_staticVariables = value;
		}

		// ----------------------------
		// extendsClass
		// ----------------------------

		private var _extendsClasses:Array /* of String */;

		/**
		 * @return An <code>Array</code> of fully qualified class names that represent the inheritance order of the current <code>Type</code>.
		 * <p>The first item in the <code>Array</code> is the fully qualified classname of the super class of the current <code>Type</code>.</p>
		 */
		public function get extendsClasses():Array {
			return _extendsClasses;
		}

		/**
		 * @private
		 */
		public function set extendsClasses(value:Array):void {
			_extendsClasses = value;
		}

		// ----------------------------
		// variables
		// ----------------------------

		private var _variables:Array;

		/**
		 * An array of <code>IMember</code> instances that describe the variables of the <code>Type</code>
		 * @see org.as3commons.reflect.IMember IMember
		 */
		public function get variables():Array {
			return _variables;
		}

		/**
		 * @private
		 */
		public function set variables(value:Array):void {
			_variables = value;
		}

		// ----------------------------
		// fields
		// ----------------------------

		private var _fields:HashArray;

		/**
		 * An array of all the static constants, constants, static variables and variables of the <code>Type</code>
		 * @see org.as3commons.reflect.IMember IMember
		 */
		public function get fields():Array {
			if (_fields == null) {
				createFieldsHashArray();
			}
			return _fields.getArray();
		}

		private function createFieldsHashArray():void {
			_fields = new HashArray(MEMBER_PROPERTY_NAME, false, accessors.concat(staticConstants).concat(constants).concat(staticVariables).concat(variables));
		}

		/**
		 * An array of Field containing all accessors and variables for the type.
		 *
		 * @see org.as3commons.reflect.Variable
		 * @see org.as3commons.reflect.Accessor
		 * @see org.as3commons.reflect.Field
		 */
		public function get properties():Array {
			return accessors.concat(variables);
		}

		// --------------------------------------------------------------------
		//
		// Public Methods
		//
		// --------------------------------------------------------------------

		/**
		 * Returns the <code>Method</code> object for the method in this type
		 * with the given name.
		 *
		 * @param name the name of the method
		 */
		public function getMethod(name:String):Method {
			return _methods.get(name);
		}

		/**
		 * Returns the <code>Field</code> object for the field in this type
		 * with the given name.
		 *
		 * @param name the name of the field
		 */
		public function getField(name:String):Field {
			//force _fields HashArray to be created
			if (_fields == null) {
				createFieldsHashArray();
			}
			return _fields.get(name);
		}

		private var _metaDataLookup:Dictionary;

		public function createMetaDataLookup():void {
			_metaDataLookup = new Dictionary(true);
			addToMetaDataLookup(_methods.getArray());
			if (_fields == null) {
				createFieldsHashArray();
			}
			addToMetaDataLookup(_fields.getArray());
		}

		/**
		 * Retrieves an Array of <code>MetadataContainer</code> instances that are associated with the current <code>Type</code>
		 * that contain <code>MetaData</code> with the specified name.
		 * <p>Each <code>MetadataContainer</code> instance can be one of these subclasses:</p>
		 * <ul>
		 *   <li>Type</li>
		 *   <li>Method</li>
		 *   <li>Accessor</li>
		 *   <li>Constant</li>
		 *   <li>Variable</li>
		 * </ul>
		 * @param name The specified <code>MetaData</code> name.
		 * @return An Array of <code>MetadataContainer</code> instances.
		 */
		public function getMetaDataContainers(name:String):Array {
			if (_metaDataLookup != null) {
				return _metaDataLookup[name] as Array;
			} else {
				return null;
			}
		}

		private function addToMetaDataLookup(containerList:Array):void {
			for each (var container:MetaDataContainer in containerList) {
				var metadatas:Array = container.metaData;
				for each (var m:MetaData in metadatas) {
					var arr:Array = _metaDataLookup[m.name];
					if (arr == null) {
						arr = [];
						_metaDataLookup[m.name] = arr;
					}
					arr[arr.length] = container;
				}
			}
		}

	}
}