/**
 * Copyright (C) 2008 Darshan Sawardekar.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
 
package org.puremvc.as3.multicore.utilities.fabrication.patterns.command {
	import flash.utils.getQualifiedClassName;
	
	import org.as3commons.reflect.Type;
	import org.puremvc.as3.multicore.interfaces.ICommand;
	import org.puremvc.as3.multicore.interfaces.IMediator;
	import org.puremvc.as3.multicore.interfaces.INotification;
	import org.puremvc.as3.multicore.interfaces.IProxy;
	import org.puremvc.as3.multicore.patterns.command.SimpleCommand;
	import org.puremvc.as3.multicore.utilities.fabrication.Fabrication;
	import org.puremvc.as3.multicore.utilities.fabrication.global.FabricationConfig;
	import org.puremvc.as3.multicore.utilities.fabrication.injection.InjectProcessorFactory;
	import org.puremvc.as3.multicore.utilities.fabrication.injection.MediatorInjectorProcessor;
	import org.puremvc.as3.multicore.utilities.fabrication.injection.ProxyInjectorProcessor;
	import org.puremvc.as3.multicore.utilities.fabrication.interfaces.IClassInfoType;
	import org.puremvc.as3.multicore.utilities.fabrication.interfaces.ICommandProcessor;
	import org.puremvc.as3.multicore.utilities.fabrication.interfaces.IDisposable;
	import org.puremvc.as3.multicore.utilities.fabrication.interfaces.IFabrication;
	import org.puremvc.as3.multicore.utilities.fabrication.interfaces.IRouter;
	import org.puremvc.as3.multicore.utilities.fabrication.patterns.facade.FabricationFacade;

	/**
	 * SimpleFabricationCommand is the base class for all fabrication commands.
	 * It aliases common facade.<methodName> invocations and provides
	 * easier access to properties often required by a fabrication application.
	 * 
	 * @author Darshan Sawardekar
	 */
	public class SimpleFabricationCommand extends SimpleCommand implements ICommandProcessor, IDisposable , IClassInfoType {

		/**
		 * Names of injected properites
		 */
		protected var injectionFieldsNames:Array;
		
		/**
		 * @see org.puremvc.as3.multicore.utilities.fabrication.interfaces#executeCommand()
		 */
		public function executeCommand(clazz:Class, body:Object = null, note:INotification = null):ICommand {
			
			return fabFacade.executeCommandClass(clazz, body, note);
		}
		
		/**
		 * Fulfill the use-case initiated by the given <code>INotification</code>.
		 * 
		 * <P>
		 * In the Command Pattern, an application use-case typically
		 * begins with some user action, which results in an <code>INotification</code> being broadcast, which 
		 * is handled by business logic in the <code>execute</code> method of an
		 * <code>ICommand</code>.</P>
		 * 
		 * @param notification the <code>INotification</code> to handle.
		 */
		override public function execute(notification:INotification):void
		{
			super.execute(notification);
			
			performInjections();
		}
		
		/**
		 * @see org.puremvc.as3.multicore.utilities.fabrication.interfaces.IDisposable#dispose()
		 */
		public function dispose():void {
			if (injectionFieldsNames) {
				var injectedFieldsNum:uint = injectionFieldsNames.length;
				for (var i:int = 0; i < injectedFieldsNum; i++) {
					
					var fieldName:String = ""+injectionFieldsNames[i];
					this[ fieldName ] = null;
				}
				injectionFieldsNames = null;
			}
			
			removeClassInfo()
		}
		
		/**
		 * Reference to the fabrication facade
		 */
		public function get fabFacade():FabricationFacade {
			return facade as FabricationFacade;
		}

		/**
		 * Reference to the main fabrication application class
		 */
		public function get fabrication():IFabrication {
			return fabFacade.getFabrication();
		}

		/**
		 * Reference to the current message router
		 */
		public function get applicationRouter():IRouter {
			return fabrication.router;
		}

		/**
		 * Alias to facade.registerCommand
		 * 
		 * @see org.puremvc.as3.multicore.interfaces.IFacade#registerCommand
		 */
		public function registerCommand( notificationName:String, commandClassRef:Class ):void {
			
			facade.registerCommand(notificationName, commandClassRef);
		}

		/**
		 * Alias to facade.removeCommand
		 * 
		 * @see org.puremvc.as3.multicore.interfaces.IFacade#removeCommand
		 */
		public function removeCommand( notificationName:String ):void {
			facade.removeCommand(notificationName);
		}

		/**
		 * Alias to facade.hasCommand
		 * 
		 * @see org.puremvc.as3.multicore.interfaces.IFacade#hasCommand
		 */
		public function hasCommand( notificationName:String ):Boolean {
			return facade.hasCommand(notificationName);
		}

		/**
		 * Alias to facade.registerProxy
		 * 
		 * @return The proxy being registered
		 * @see org.puremvc.as3.multicore.interfaces.IFacade#registerProxy
		 */
		public function registerProxy( proxy:IProxy ):IProxy {
			if(hasProxy(proxy.getProxyName())){
				throw new Error("already have proxy: " + proxy.getProxyName());
			}
			facade.registerProxy(proxy);
			return proxy;	
		}

		/**
		 * Alias to facade.retrieveProxy
		 * 
		 * @see org.puremvc.as3.multicore.interfaces.IFacade#retrieveProxy
		 */
		public function retrieveProxy( proxyName:String ):IProxy {
			return facade.retrieveProxy(proxyName);	
		}

		/**
		 * Alias to facade.removeProxy
		 * 
		 * @see org.puremvc.as3.multicore.interfaces.IFacade#removeProxy
		 */
		public function removeProxy( proxyName:String ):IProxy {
			return facade.removeProxy(proxyName);
		}

		/**
		 * Alias to facade.hasProxy
		 * 
		 * @see org.puremvc.as3.multicore.interfaces.IFacade#hasProxy
		 */
		public function hasProxy( proxyName:String ):Boolean {
			return facade.hasProxy(proxyName);
		}

		/**
		 * Alias to facade.registerMediator
		 * 
		 * @return The mediator being registered
		 * @see org.puremvc.as3.multicore.interfaces.IFacade#registerMediator
		 */
		public function registerMediator( mediator:IMediator ):IMediator {
			facade.registerMediator(mediator);
			return mediator;
		}

		/**
		 * Alias to facade.retrieveMediator
		 * 
		 * @see org.puremvc.as3.multicore.interfaces.IFacade#retrieveMediator
		 */
		public function retrieveMediator( mediatorName:String ):IMediator {
			return facade.retrieveMediator(mediatorName) as IMediator;
		}

		/**
		 * Alias to facade.removeMediator
		 * 
		 * @see org.puremvc.as3.multicore.interfaces.IFacade#removeMediator
		 */
		public function removeMediator( mediatorName:String ):IMediator {
			return facade.removeMediator(mediatorName);
		}

		/**
		 * Alias to facade.hasMediator
		 * 
		 * @see org.puremvc.as3.multicore.interfaces.IFacade#hasMediator
		 */
		public function hasMediator( mediatorName:String ):Boolean {
			return facade.hasMediator(mediatorName);
		}

		/**
		 * Alias to fabFacade.routeNotification
		 * 
		 * @see org.puremvc.as3.multicore.utilities.fabrication.patterns.facade.FabricationFacade#routeNotification
		 */
		override public function routeNotification(noteName:Object, noteBody:Object = null, noteType:String = null, to:Object = null):void {
			fabFacade.routeNotification(noteName, noteBody, noteType, to);
		}

		/**
		 * Alias to fabFacade.registerInterceptor
		 * 
		 * @see org.puremvc.as3.multicore.utilities.fabrication.patterns.facade.FabricationFacade#registerInterceptor
		 */
		public function registerInterceptor(noteName:String, clazz:Class, parameters:Object = null):void {
			
			fabFacade.registerInterceptor(noteName, clazz, parameters);
		}

		/**
		 * Alias to fabFacade.removeInterceptor
		 * 
		 * @see org.puremvc.as3.multicore.utilities.fabrication.patterns.facade.FabricationFacade#removeInterceptor
		 */
		public function removeInterceptor(noteName:String, clazz:Class = null):void {
			fabFacade.removeInterceptor(noteName, clazz);
		}
		
		/**
		 * Performs injection action on current FlexMediator.
		 * By convention we allow proxies and mediators injection on
		 * FlexMediator instance
		 * @see org.puremvc.as3.multicore.utilities.fabrication.injection.ProxyInjector
		 * @see org.puremvc.as3.multicore.utilities.fabrication.injection.MediatorInjector
		 */
		protected function performInjections():void
		{
			if(injectionFieldsNames != null) return ;
			
			injectionFieldsNames = [];
			injectionFieldsNames = InjectProcessorFactory.getInstance().inject(facade,this)
		}
		
		private var classInfoType:Type;
		
		public function getClassInfo():Type
		{
			if(classInfoType != null) return classInfoType;
			classInfoType = Type.forInstance(this);
			return classInfoType;
		}
		
		public function removeClassInfo():void
		{
			if(classInfoType!=null){
				classInfoType.dispose();
			}
			classInfoType = null
		}
		
		/**
		 * get inject name :
		 * sample: resProxy
		 */ 
		public function haveInjectName(nm:String):Boolean
		{
			return this[nm] != null
		}
		
		public function setInjectName(nm:String,obj:Object):void
		{
			this[nm] = obj
		}
	}
}
